import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ['resnet', 'ResNet18']

class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion * planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion * planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, width=16, num_classes=10, n_channel=3):
        super(ResNet, self).__init__()
        self.in_planes = width

        self.conv1 = nn.Conv2d(n_channel, width, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(width)
        self.layer1 = self._make_layer(block, width, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, width * 2, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, width * 4, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, width * 8, num_blocks[3], stride=2)
        self.linear = nn.Linear(width * 8 * block.expansion, num_classes)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out

def resnet(depth=18, width=16, num_classes=10, n_channel=3):
    # custom resnet
    # width=16 in original ResNet18 in He et al.
    if depth == 18:
        block = BasicBlock
        num_blocks = [2, 2, 2, 2]
    elif depth == 34:
        block = BasicBlock
        num_blocks = [3, 4, 6, 3]
    elif depth == 50:
        block = Bottleneck
        num_blocks = [3, 4, 6, 3]
    elif depth == 101:
        block = Bottleneck
        num_blocks = [3, 4, 23, 3]
    elif depth == 152:
        block = Bottleneck
        num_blocks = [3, 8, 36, 3]
    else:
        raise KeyError('Depth %i is not defined for resnet' % depth)
    return ResNet(block, num_blocks=num_blocks, width=width, num_classes=num_classes, n_channel=n_channel)

def ResNet18(num_classes=10, n_channel=3):
    # ResNet18 commonly seen in adversarial training
    return ResNet(BasicBlock, [2, 2, 2, 2], width=64, num_classes=num_classes, n_channel=n_channel)

def ResNet34():
    return ResNet(BasicBlock, [3, 4, 6, 3], width=64)

def ResNet50():
    return ResNet(Bottleneck, [3, 4, 6, 3], width=64)

def ResNet101():
    return ResNet(Bottleneck, [3, 4, 23, 3], width=64)

def ResNet152():
    return ResNet(Bottleneck, [3, 8, 36, 3], width=64)


def test():
    # net = ResNet18()
    net = resnet(depth=18, width=16)
    y = net(torch.randn(1, 3, 32, 32))
    print(y.size())

if __name__ == '__main__':
    test()
